#
# Copyright (C) 2013,2014,2016  Stuart D. Gathman (stuart@gathman.org)
# Summary:	Python ANSI EDI file manipulation
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
#   This program is distributed in the hope that it will be useful, but WITHOUT
#   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#   more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import copy
import StringIO
import datetime
import time
import decimal
import edicc

class EDIException(Exception):
  pass

def pad(s,minlen,maxlen=None):
  if not maxlen: maxlen = minlen
  if s is None:
    s = ''
  else:
    if len(s) > maxlen:
      raise EDIException("ANSI element overflow: "+s)
    if len(s) >= minlen:
      return s
  while len(s) < minlen:
    s += ' '
  return s

def splitBatch(msg):
  while msg:
    if not msg.startswith('ISA') or len(msg) < 177:
      raise ValueError('not ANSI X12')
    segsep = msg[105]
    pos = msg.find(segsep+'IEA')
    if pos < 0:
      raise ValueError('missing IEA segment')
    pos = msg.find(segsep,pos+4)
    if pos < 0:
      raise ValueError('IEA segment missing terminator')
    pos += 1
    m,msg = msg[:pos],msg[pos:]
    yield m

## EDI segment
# A segment is an auto-extending 2D variable array.  
class EDISeg(object):
  def __init__(self,msg,seg=[]):
    if type(seg) is list:
      self.seg = seg
      self.s = None
    else:
      self.seg = None
      self.s = seg
    self.msg = msg

  def get(self,i,sub=None):
    if self.seg is None:
      self.seg = self.msg.strtoseg(self.s)
    try:
      s = self.seg[i]
    except: s = ''
    if sub is not None:
      if type(s) is not list:
        if sub == 0: return s
      elif sub < len(s):
        return s[sub]
      return ''
    return s

  def put(self,s,i,sub=None):
    if self.seg is None:
      self.seg = self.msg.strtoseg(self.s)
    else:
      assert type(self.seg) is list
    seg = self.seg
    self.s = None
    while len(seg) <= i:
      seg.append('')
    if sub is not None:
      a = seg[i]
      if type(a) is not list:
        a = [a]
      while len(a) <= sub:
        a.append('')
      a[sub] = s
      seg[i] = a
    else:
      seg[i] = s

  def __str__(self):
    if self.s is None:
      msg = self.msg
      self.s = msg.segtostr(msg.elemtostr(e) for e in self.seg)
    return self.s

  def getstr(self,i,sub=None):
    return self.msg.getstr(self,i,sub)

  def putstr(self,s,i,sub=None):
    self.msg.putstr(s,i,sub,self)

  def clear(self):
    self.s = None
    self.seg = []

  def append(self,s):
    if self.seg is None:
      self.seg = self.msg.strtoseg(self.s)
    self.seg.append(s)

  def __iter__(self):
    if self.seg is None:
      self.seg = self.msg.strtoseg(self.s)
    return iter(self.seg)

class ANSIMsg(object):
  def __init__(self,segin=None):
    self.setsep('*',':','\n')
    self.authqual = '00'
    self.authinfo = ''
    self.secqual = '00'
    self.secinfo = ''
    self.isasendqual = 'ZZ'
    self.isasender = None
    self.isarecvqual = 'ZZ'
    self.isareceiver = None
    self.standard = 'U'
    self.isaversion = '00401'
    self.isactlno = None
    self.ackreq = '0'
    self.testflag = 'P'
    self.func = None
    self.agency = 'X'
    self.version = '004010'
    self.sender = None
    self.recipient = None
    self.ctlno = None
    self.transet = None
    self.tranid = None
    self._nexttran = 1000
    self._nextgroup = 1000
    self.segcnt = 0
    self.curiter = None
    self.clear(segin)

  ## Generate a new transaction control no.
  # It is required to be unique within a GS/GE group, and the 
  # default behaviour is to be unique for the lifetime of this
  # ANSIMsg instance, but it can be useful to make it globally unique.
  def nextTran(self):
    self._nexttran += 1
    return self._nexttran

  def nextGroup(self):
    self._nextgroup += 1
    return self._nextgroup

  GROUPS = {
    753: 'RR',
    754: 'RG',
    810: 'IN',
    824: 'AG',
    850: 'PO',
    855: 'PR',
    856: 'SH',
    857: 'BS',
    860: 'PC',
    997: 'FA'
  }
  def getGroup(self,transet):
    return self.GROUPS.get(int(transet),None)

  def clear(self,segin=None):
    if segin:
      self.segs = self.itersegs(segin)
    else:
      self.segs = []
    self.curseg = EDISeg(self)

  def addElem(self,s):
    self.curseg.append(s)

  ## Replace separator chars in a string, and map to allowed charset.
  def sanitize(self,s):
    s = str(s)
    s = s.encode('ascii','replace')
    return s.translate(None,self._sep)

  def setsep(self,elem=None,sub=None,seg=None):
    if elem: self._elemsep = elem
    if sub: self._subsep = sub
    if seg: self._segsep = seg
    if self._elemsep == self._segsep[0]:
      raise ValueError('ANSI separators not unique')
    self._sep = self._elemsep + self._subsep + self._segsep

  def elemtostr(self,e):
    if type(e) is not list:
      return self.sanitize(e)
    return self._subsep.join(self.sanitize(s) for s in e)

  def segtostr(self,seg):
    return self._elemsep.join(seg)

  def strtoelem(self,s):
    a = s.split(self._subsep)
    if len(a) == 1: return a[0]
    return a

  def strtoseg(self,s):
    return [self.strtoelem(e) for e in s.split(self._elemsep)]

  def append(self,s):
    if isinstance(s,EDISeg):
      # copy EDISeg to str since EDISeg is mutable
      # use *our* separators since EDISeg might belong to another ANSIMsg
      s = self.segtostr(self.elemtostr(e) for e in s)
      # FIXME: parse str entries in self.segs when separators change
    else:
      s = str(s)
    self.segs.append(s)
    self.segcnt += 1

  def write(self):
    #self.segs.append(self.curseg)
    #self.curseg = EDISeg(self)
    self.append(self.curseg)
    self.curseg.clear()

  def writeto(self,fp):
    # self.segs can be a mixture of EDISeg and str
    for s in self.segs:
      fp.write(str(s)+self._segsep)

  def __str__(self):
    fp = StringIO.StringIO()
    self.writeto(fp)
    return fp.getvalue()

  def ignore(self,s):
    while self.curiter and self.getstr(0) == s:
      self.next()

  ## Return a sequence of EDI segments. 
  # @param s input message as string
  def itersegs(self,s):
    while s:
      if s.startswith('ISA'):
        elemsep = s[3]
        fld = [3]
        try:
          while len(fld) < 16:
            fld.append(s.index(elemsep,fld[-1]+1))
        except ValueError:
          raise ValueError('not ANSI X12')
        seppos = fld[-3]
        pos = s.find('GS',seppos)
        if pos < 0:
          pos = s.find('ST',seppos)
        if pos < 0:
          raise ValueError('missing GS or ST segment')
        seppos = s.rindex(elemsep,seppos,pos)
        self.setsep(s[seppos],s[seppos+1],s[seppos+2:pos])
      pos = s.find(self._segsep[0])
      if pos < 0: return
      seg,s = s[:pos],s[pos+1:].lstrip()
      yield seg

  def __iter__(self):
    if not self.curiter:
      def itersegs():
        for s in self.segs:
          seg = EDISeg(self,self.strtoseg(s))
          self.curseg = seg
          elem0 = seg.get(0)
          if elem0 == 'ISA':
            self.authqual = seg.get(1)
            self.authinfo = seg.get(2)
            self.secqual = seg.get(3)
            self.secinfo = seg.get(4)
            self.isasendqual = seg.get(5)
            self.isasender = seg.get(6)
            self.isarecvqual = seg.get(7)
            self.isareceiver = seg.get(8)
            #self.isatimesent = seg.getdatetime(9,10,False)
            self.standard = seg.get(11)
            self.isaversion = seg.get(12)
            self.isactlno = seg.get(13)
            self.ackreq = seg.get(14)
            self.testflag = seg.get(15)
          elif elem0 == 'GS':
            self.func = seg.get(1)
            self.sender = seg.get(2)
            self.recipient = seg.get(3)
            #self.timesent = seg.getdatetime(4,5)
            self.ctlno = seg.get(6)
            self.agency = seg.get(7)
            self.version = seg.get(8)
            self.errs = 0
            self.accepted = 0
          elif elem0 == 'ST':
            self.transet = seg.get(1)
            self.tranid = seg.get(2)
          yield seg
      self.curiter = itersegs()
    return self.curiter

  def next(self):
    iter(self).next()

  def reset(self):
    self.curiter = None

  def startISA(self,ctlno=None):
    if ctlno:
      self.isactlno = ctlno
    now = datetime.datetime.now()
    self.putstr('ISA',0)
    self.putstr(pad(self.authqual,2),1)
    self.putstr(pad(self.authinfo,10),2)
    self.putstr(pad(self.secqual,2),3)
    self.putstr(pad(self.secinfo,10),4)
    self.putstr(pad(self.isasendqual,2),5)
    self.putstr(pad(self.isasender,15),6)
    self.putstr(pad(self.isarecvqual,2),7)
    self.putstr(pad(self.isareceiver,15),8)
    self.putdatetime(now,9,10,century=False)
    self.putstr(self.standard,11)
    self.putstr(pad(self.isaversion,5),12)
    self.putstr(self.isactlno,13)
    self.putstr(self.ackreq,14)
    self.putstr(self.testflag,15)
    self.putstr('',16,1)
    self.write()
    self.groupcnt = 0

  def startGS(self,ctlno=None):
    if ctlno:
      self.ctlno = ctlno
    else:
      self.ctlno = self.nextGroup()
    now = datetime.datetime.now()
    self.putstr('GS',0)
    self.putstr(self.func,1)
    self.putstr(self.sender,2)
    self.putstr(self.recipient,3)
    self.putdatetime(now,4,5)
    self.putstr(self.ctlno,6)
    self.putstr(self.agency,7)
    self.putstr(self.version,8)
    self.write()
    self.trancnt = 0

  def closeGS(self):
    if not self.ctlno: return
    self.close()
    self.putstr('GE',0)
    self.putnum(self.trancnt,1)
    self.putstr(self.ctlno,2)
    self.write()
    self.groupcnt += 1
    self.ctlno = None

  def closeISA(self):
    if not self.isactlno: return
    self.closeGS()
    self.putstr('IEA',0)
    self.putnum(self.groupcnt,1)
    self.putstr(self.isactlno,2)
    self.write()
    self.isactlno = None

  def getstr(self,i,sub=None,seg=None):
    if not seg: seg = self.curseg
    return seg.get(i,sub)

  def getnum(self,i,sub=None,seg=None):
    if not seg: seg = self.curseg
    s = seg.get(i,sub)
    if '.' in s:
      return decimal.Decimal(s)
    return int(s)

  def putstr(self,s,i,sub=None,seg=None):
    if not seg: seg = self.curseg
    seg.put(s,i,sub)

  def putnum(self,n,i,fix=0,sub=None,seg=None):
    if not seg: seg = self.curseg
    if fix != 0:
      raise NotImplementedException()
    seg.put(str(n),i,sub)

  def getdate(self,i,sub=None,seg=None):
    if not seg: seg = self.curseg
    d = int(seg.get(i,sub))
    y = d/10000
    d %= 10000
    return datetime.date(y,d/100,d%100)

  def putdate(self,dt,i,seg=None,century=True):
    if not seg: seg = self.curseg
    if century:
      fmt = '%Y%m%d'
    else:
      fmt = '%y%m%d'
    if type(dt) is float:
        t = time.gmtime(dt)
        seg.put(time.strftime(fmt,t),i)
    else:
        seg.put(dt.strftime(fmt),i)

  def putdatetime(self,dt,f1,f2,seg=None,century=True):
    if not seg: seg = self.curseg
    if type(dt) is float:
        t = time.gmtime(dt)
        self.putdate(t,f1,seg,century)
        seg.put(time.strftime('%H%M',t),f2)
    else:
        self.putdate(dt,f1,seg,century)
        seg.put(dt.strftime('%H%M'),f2)

  def open(self,transet,tranid=None):
    self.close()
    func = self.getGroup(transet)
    if not func:
      raise ValueError('Unknown transet: '+transet)
    if func != self.func:
      self.closeGS()
      self.func = func
      self.startGS()
    self.transet = transet
    if tranid:
      self.tranid = tranid
    else:
      self.tranid = self.nextTran()
    self.putstr('ST',0)
    self.putstr(self.transet,1)
    s = str(self.tranid)
    while len(s) < 4: s = '0' + s
    self.putstr(s,2)
    self.segcnt = 0
    self.write()

  def is_open(self): return not not self.tranid

  def close(self):
    if not self.tranid: return
    self.putstr('SE',0)
    self.putnum(self.segcnt+1,1)
    self.putstr(self.tranid,2)
    self.write()
    self.tranid = None
    self.trancnt += 1

class ANSIReceive(ANSIMsg):
  
  def __init__(self):
    ANSIMsg.__init__(self)
    self.acks = {}
    self.noacks = 0
    self.msg = None

  def startAck(self,msg):
    self.closeAck()
    assert msg.getstr(0) == 'GS'
    self.acks = {}
    self.noacks = 0
    self.sender = msg.recipient
    self.recipient = msg.sender
    self.msg = msg

  def acktran(self,code='A',reason=None):
    if not self.msg: return
    msg = self.msg
    key = msg.transet,msg.tranid
    if key not in self.acks:
      self.acks[key] = code,[]
    c,reasons = self.acks[key]
    if reason:
      reasons.append(reason)
    if code > c:
      self.acks[key] = code,reasons

  ## Reject transaction for syntax reason.
  def reject(self,reason=None):
    self.acktran('R',reason)

  ## Cancel ack for transaction (to be acked elsewhere)
  def cancel(self):
    if not self.msg: return
    msg = self.msg
    key = msg.transet,msg.tranid
    del self.acks[key]
    self.noacks += 1

  def warn(self,reason):
    self.acktran('E',reason)

  def closeAck(self):
    if not self.msg: return
    msg = self.msg
    if msg.tranid:
      self.reject('3')
    sent = 0
    ctlno = 0
    grpctlno = None
    if msg.getstr(0) == 'GE':
      try:
        sent = int(msg.getstr(1))
        ctlno = int(msg.getstr(2))
        grpctlno = int(msg.ctlno)
      except ValueError: pass
    trancnt = len(self.acks) + self.noacks
    self.msg = None
    if sent and sent == self.noacks: return
    accepted = len([c for c,r in self.acks.itervalues() if c == 'A'])
    self.open('997')
    self.putstr('AK1',0)
    self.putstr(msg.func,1)
    self.putnum(grpctlno,2)
    self.write()
    if trancnt != accepted:
      a = list(self.acks.iteritems())
      a.sort()
      for key,val in a:
        transet,tranid = key
        code,reasons = val
        self.putstr('AK2',0)
        self.putstr(transet,1)
        self.putstr(tranid,1)
        self.write()
        self.putstr('AK5',0)
        self.putstr(code,1)
        for reason in reasons:
          self.addElem(reason)
        self.write()
    self.putstr('AK9',0)
    self.putnum(sent,2)        # number reported sent - 0 for no trailer
    self.putnum(trancnt,3)     # received
    self.putnum(accepted,4)    # accepted
    if accepted == 0:
      self.putstr('R',1) # rejected all
    elif trancnt != sent or ctlno != grpctlno:
      self.putstr('E',1) # errors noted
    elif trancnt == accepted:
      self.putstr('A',1) # all accepted
    else:
      self.putstr('P',1) # partial - some accepted
    # append reason codes
    if not sent:
      self.addElem('3')       # missing trailer
    else:
      if trancnt != sent:
        self.addElem('5')       # trancnt doesn't match
      if ctlno != grpctlno:
        self.addElem('4')       # ctlno doesn't match
    self.write()
    self.close()

  def iter_trans(self,msg):
    for seg in msg:
      elem0 = seg.get(0)
      if elem0 == 'GS':
        self.closeAck()
        submsg = copy.copy(msg)
        if msg.func != 'FA':
          self.startAck(msg)
      elif elem0 == 'ST':
        submsg.clear()
        submsg.transet = msg.transet
        submsg.tranid = msg.tranid
        segcnt = 1
      elif elem0 == 'SE':
        segcnt += 1
        cnt = 0
        tranid = None
        try:
          cnt = msg.getnum(1)
          tranid = msg.getstr(2)
        except ValueError: pass
        self.acktran()
        if cnt != segcnt:
          self.reject('5')
        if tranid != msg.tranid:
          self.reject('4')
        yield submsg
        msg.tranid = None
        msg.transet = None
      elif elem0 == 'GE':
        self.closeAck()
      elif elem0 in ('ISA','IEA'):
        self.closeAck()
      else:
        segcnt += 1
        submsg.append(seg)
